@@ -110,6 +110,7 @@ class Command(object):
         self._sdr = None
         self._oem = None
         self._netchannel = None
+        self._ipv6support = None
         if onlogon is not None:
             self.ipmi_session = session.Session(bmc=bmc,
                                                 userid=userid,
@@ -127,8 +128,27 @@ class Command(object):
                                                 port=port,
                                                 kg=kg)
 
+    def register_key_handler(self, callback, type='tls'):
+        """Assign a verification handler for a public key
+
+        When the library attempts to communicate with the management target
+        using a non-IPMI protocol, it will try to verify a key.  This
+        allows a caller to register a key handler for accepting or rejecting
+        a public key/certificate.  The callback will be passed the peer public
+        key or certificate.
+
+        :param callback:  The function to call with public key/certificate
+        :param type: Whether the callback is meant to handle 'tls' or 'ssh',
+                     defaults to 'tls'
+        """
+        if type == 'tls':
+            self._certverify = callback
+        self.oem_init()
+        self._oem.register_key_handler(callback, type)
+
     def logged(self, response):
         self.onlogon(response, self)
+        self.onlogon = None
 
     @classmethod
     def eventloop(cls):
@@ -145,6 +165,19 @@ class Command(object):
         """
         return session.Session.wait_for_rsp(timeout=timeout)
 
+    def _get_device_id(self):
+        response = self.raw_command(netfn=0x06, command=0x01)
+        if 'error' in response:
+            raise exc.IpmiException(response['error'], code=response['code'])
+        return {
+            'device_id': response['data'][0],
+            'device_revision': response['data'][1] & 0b1111,
+            'manufacturer_id': struct.unpack(
+                '<I', struct.pack('3B', *response['data'][6:9]) + '\x00')[0],
+            'product_id': struct.unpack(
+                '<H', struct.pack('2B', *response['data'][9:11]))[0],
+        }
+
     def oem_init(self):
         """Initialize the command object for OEM capabilities
 
@@ -155,17 +188,7 @@ class Command(object):
         """
         if self._oem:
             return
-        response = self.raw_command(netfn=6, command=1)
-        if 'error' in response:
-            raise exc.IpmiException(response['error'], code=response['code'])
-        self._oem = get_oem_handler({
-            'device_id': response['data'][0],
-            'device_revision': response['data'][1] & 0b1111,
-            'manufacturer_id': struct.unpack(
-                '<I', struct.pack('3B', *response['data'][6:9]) + '\x00')[0],
-            'product_id': struct.unpack(
-                '<H', struct.pack('2B', *response['data'][9:11]))[0],
-        }, self)
+        self._oem = get_oem_handler(self._get_device_id(), self)
 
     def get_bootdev(self):
         """Get current boot device override information.
@@ -267,6 +290,16 @@ class Command(object):
         else:
             return lastresponse
 
+    def get_video_launchdata(self):
+        """Get data required to launch a remote video session to target.
+
+        This is a highly proprietary scenario, the return data may vary greatly
+        host to host.  The return should be a dict describing the type of data
+        and the data.  For example {'jnlp': jnlpstring}
+        """
+        self.oem_init()
+        return self._oem.get_video_launchdata()
+
     def reset_bmc(self):
         """Do a cold reset in BMC
         """
@@ -322,7 +355,7 @@ class Command(object):
         return {'bootdev': bootdev}
 
     def xraw_command(self, netfn, command, bridge_request=(), data=(),
-                     delay_xmit=None):
+                     delay_xmit=None, retry=True, timeout=None):
         """Send raw ipmi command to BMC, raising exception on error
 
         This is identical to raw_command, except it raises exceptions
@@ -336,18 +369,23 @@ class Command(object):
         :param bridge_request: The target slave address and channel number for
                                the bridge request.
         :param data: Command data as a tuple or list
+        :param retry: Whether to retry this particular payload or not, defaults
+                      to true.
+        :param timeout: A custom time to wait for initial reply, useful for
+                        a slow command.  This may interfere with retry logic.
         :returns: dict -- The response from IPMI device
         """
         rsp = self.ipmi_session.raw_command(netfn=netfn, command=command,
                                             bridge_request=bridge_request,
-                                            data=data, delay_xmit=delay_xmit)
+                                            data=data, delay_xmit=delay_xmit,
+                                            retry=retry, timeout=timeout)
         if 'error' in rsp:
             raise exc.IpmiException(rsp['error'], rsp['code'])
         rsp['data'] = buffer(bytearray(rsp['data']))
         return rsp
 
     def raw_command(self, netfn, command, bridge_request=(), data=(),
-                    delay_xmit=None):
+                    delay_xmit=None, retry=True, timeout=None):
         """Send raw ipmi command to BMC
 
         This allows arbitrary IPMI bytes to be issued.  This is commonly used
@@ -360,11 +398,15 @@ class Command(object):
         :param bridge_request: The target slave address and channel number for
                                the bridge request.
         :param data: Command data as a tuple or list
+        :param retry: Whether or not to retry command if no response received.
+                      Defaults to True
+        :param timeout: A custom amount of time to wait for initial reply
         :returns: dict -- The response from IPMI device
         """
         return self.ipmi_session.raw_command(netfn=netfn, command=command,
                                              bridge_request=bridge_request,
-                                             data=data, delay_xmit=delay_xmit)
+                                             data=data, delay_xmit=delay_xmit,
+                                             retry=retry, timeout=timeout)
 
     def get_power(self):
         """Get current power state of the managed system
@@ -512,8 +554,15 @@ class Command(object):
         if 'error' not in guiddata:
             zerofru['UUID'] = pygutil.decode_wireformat_uuid(
                 guiddata['data'])
-        if not zerofru:
-            zerofru = None
+        # Add some fields returned by get device ID command to FRU 0
+        # Also rename them to something more in line with FRU 0 field naming
+        # standards
+        device_id = self._get_device_id()
+        device_id['Device ID'] = device_id.pop('device_id')
+        device_id['Device Revision'] = device_id.pop('device_revision')
+        device_id['Manufacturer ID'] = device_id.pop('manufacturer_id')
+        device_id['Product ID'] = device_id.pop('product_id')
+        zerofru.update(device_id)
         return self._oem.process_fru(zerofru)
 
     def get_inventory(self):
@@ -669,12 +718,13 @@ class Command(object):
             cmddata = bytearray((channel, 12)) + socket.inet_aton(ipv4_gateway)
             self.xraw_command(netfn=0xc, command=1, data=cmddata)
 
-    def get_net_configuration(self, channel=None):
+    def get_net_configuration(self, channel=None, gateway_macs=True):
         """Get network configuration data
 
         Retrieve network configuration from the target
 
         :param channel: Channel to configure, defaults to None for 'autodetect'
+        :param gateway_macs: Whether to retrieve mac addresses for gateways
         :returns: A dictionary of network configuration data
         """
         if channel is None:
@@ -697,10 +747,13 @@ class Command(object):
             channel, 4)]
         retdata['mac_address'] = self._fetch_lancfg_param(channel, 5)
         retdata['ipv4_gateway'] = self._fetch_lancfg_param(channel, 12)
-        retdata['ipv4_gateway_mac'] = self._fetch_lancfg_param(channel, 13)
         retdata['ipv4_backup_gateway'] = self._fetch_lancfg_param(channel, 14)
-        retdata['ipv4_backup_gateway_mac'] = self._fetch_lancfg_param(channel,
-                                                                      15)
+        if gateway_macs:
+            retdata['ipv4_gateway_mac'] = self._fetch_lancfg_param(channel, 13)
+            retdata['ipv4_backup_gateway_mac'] = self._fetch_lancfg_param(
+                channel, 15)
+        self.oem_init()
+        self._oem.add_extra_net_configuration(retdata)
         return retdata
 
     def get_sensor_data(self):
@@ -910,6 +963,18 @@ class Command(object):
         rsp = self.xraw_command(netfn=0xc, command=2, data=(channel, 16, 0, 0))
         return rsp['data'][1:].partition('\x00')[0]
 
+    @property
+    def _supports_standard_ipv6(self):
+        # Supports the *standard* ipv6 commands for various things
+        # used to internally steer some commands to standard or OEM
+        # handler of commands
+        lanchan = self.get_network_channel()
+        if self._ipv6support is None:
+            rsp = self.raw_command(netfn=0xc, command=0x2, data=(2, lanchan,
+                                                                 0x32, 0, 0))
+            self._ipv6support = rsp['code'] == 0
+        return self._ipv6support
+
     def set_alert_destination(self, ip=None, acknowledge_required=None,
                               acknowledge_timeout=None, retries=None,
                               destination=0, channel=None):
@@ -939,10 +1004,17 @@ class Command(object):
                 destdata.extend(parsedip)
                 destdata.extend('\x00\x00\x00\x00\x00\x00')
             except socket.error:
-                parsedip = socket.inet_pton(socket.AF_INET6, ip)
-                destdata.append(0b10000000)
-                destdata.extend(parsedip)
-            self.xraw_command(netfn=0xc, command=1, data=destdata)
+                if self._supports_standard_ipv6:
+                    parsedip = socket.inet_pton(socket.AF_INET6, ip)
+                    destdata.append(0b10000000)
+                    destdata.extend(parsedip)
+                else:
+                    destdata = None
+                    self.oem_init()
+                    self._oem.set_alert_ipv6_destination(ip, destination,
+                                                         channel)
+            if destdata:
+                self.xraw_command(netfn=0xc, command=1, data=destdata)
         if (acknowledge_required is not None or retries is not None or
                 acknowledge_timeout is not None):
             currtype = self.xraw_command(netfn=0xc, command=2, data=(
@@ -1618,7 +1690,10 @@ class Command(object):
         """Retrieve OEM Firmware information
         """
         self.oem_init()
-        return self._oem.get_oem_firmware()
+        mcinfo = self.xraw_command(netfn=6, command=1)
+        bmcver = '{0}.{1}'.format(
+            ord(mcinfo['data'][2]), hex(ord(mcinfo['data'][3]))[2:])
+        return self._oem.get_oem_firmware(bmcver)
 
     def get_capping_enabled(self):
         """Get PSU based power capping status
@@ -1655,3 +1730,27 @@ class Command(object):
         """
         self.oem_init()
         self._oem.set_oem_domain_name(name)
+
+    def get_graphical_console(self):
+        """Get graphical console launcher"""
+        self.oem_init()
+        return self._oem.get_graphical_console()
+
+    def attach_remote_media(self, url, username=None, password=None):
+        """Attach remote media by url
+
+        Given a url, attach remote media (cd/usb image) to the target system.
+
+        :param url:  URL to indicate where to find image (protocol support
+                     varies by BMC)
+        :param username: Username for endpoint to use when accessing the URL.
+                         If applicable, 'domain' would be indicated by '@' or
+                         '\' syntax.
+        :param password: Password for endpoint to use when accessing the URL.
+        """
+        self.oem_init()
+        self._oem.attach_remote_media(url, username, password)
+
+    def detach_remote_media(self):
+        self.oem_init()
+        self._oem.detach_remote_media()
