@@ -124,7 +124,7 @@ class Console(object):
            #BMC tells us this is the maximum allowed size
         #data[6:7] is the promise of how small packets are going to be, but we
         #don't have any reason to worry about it
-        if (data[8] + (data[9] << 8)) != 623:
+        if (data[8] + (data[9] << 8)) not in (623, 28418):
             #TODO(jbjohnso): support atypical SOL port number
             raise NotImplementedError("Non-standard SOL Port Number")
         #ignore data[10:11] for now, the vlan detail, shouldn't matter to this
@@ -182,8 +182,13 @@ class Console(object):
         """
         self.ipmi_session.unregister_keepalive(self.keepaliveid)
         if self.activated:
-            self.ipmi_session.raw_command(netfn=6, command=0x49,
-                                          data=(1, 1, 0, 0, 0, 0))
+            try:
+                self.ipmi_session.raw_command(netfn=6, command=0x49,
+                                              data=(1, 1, 0, 0, 0, 0))
+            except exc.IpmiException:
+                # if underlying ipmi session is not working, then
+                # run with the implicit success
+                pass
 
     def send_data(self, data):
         if self.broken:
@@ -320,7 +325,11 @@ class Console(object):
             #Why not put pending data into the ack? because it's rare
             #and might be hard to decide what to do in the context of
             #retry situation
-            self.send_payload(ackpayload, retry=False)
+            try:
+                self.send_payload(ackpayload, retry=False)
+            except exc.IpmiException:
+                #if the session is broken, then close the SOL session
+                self.close()
         if self.myseq != 0 and ackseq == self.myseq:  # the bmc has something
                                                       # to say about last xmit
             self.awaitingack = False
@@ -342,9 +351,15 @@ class Console(object):
                     self._sendpendingoutput()
             if len(self.pendingoutput) > 0:
                 self._sendpendingoutput()
-        elif self.awaitingack:  # session marked us as happy, but we are not
-                #this does mean that we will occasionally retry a packet
-                #sooner than retry suggests, but that's no big deal
+        elif ackseq != 0 and self.awaitingack:
+            # if an ack packet came in, but did not match what we
+            # expected, retry our payload now.
+            # the situation that was triggered was a senseless retry
+            # when data came in while we xmitted.  In theory, a BMC
+            # should handle a retry correctly, but some do not, so
+            # try to mitigate by avoiding overeager retries
+            # occasional retry of a packet
+            # sooner than timeout suggests is evidently a big deal
             self.send_payload(payload=self.lastpayload)
 
     def main_loop(self):
