@@ -34,6 +34,7 @@ import pyghmi.constants as const
 import pyghmi.exceptions as exc
 import pyghmi.ipmi.private.constants as ipmiconst
 import struct
+import weakref
 
 TYPE_UNKNOWN = 0
 TYPE_SENSOR = 1
@@ -584,14 +585,15 @@ class SDR(object):
     :param ipmicmd: A Command class object
     """
     def __init__(self, ipmicmd):
-        self.ipmicmd = ipmicmd
+        self.ipmicmd = weakref.proxy(ipmicmd)
         self.sensors = {}
         self.fru = {}
         self.read_info()
 
     def read_info(self):
         #first, we want to know the device id
-        rsp = self.ipmicmd.raw_command(netfn=6, command=1)
+        rsp = self.ipmicmd.xraw_command(netfn=6, command=1)
+        rsp['data'] = bytearray(rsp['data'])
         self.device_id = rsp['data'][0]
         self.device_rev = rsp['data'][1] & 0b111
         # Going to ignore device available until get sdr command
@@ -618,7 +620,8 @@ class SDR(object):
         return rsp['data'][0] + (rsp['data'][1] << 8)
 
     def get_sdr(self):
-        repinfo = self.ipmicmd.raw_command(netfn=0x0a, command=0x20)
+        repinfo = self.ipmicmd.xraw_command(netfn=0x0a, command=0x20)
+        repinfo['data'] = bytearray(repinfo['data'])
         if (repinfo['data'][0] != 0x51):
             # we only understand SDR version 51h, the only version defined
             # at time of this writing
